/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.impl;

import org.apache.axis.message.addressing.Action;
import org.apache.axis.message.addressing.AddressingHeaders;
import org.apache.axis.message.addressing.util.AddressingUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.Soap1_1Constants;
import org.apache.ws.resource.JndiConstants;
import org.apache.ws.resource.Resource;
import org.apache.ws.resource.ResourceContext;
import org.apache.ws.resource.ResourceContextException;
import org.apache.ws.resource.ResourceException;
import org.apache.ws.resource.ResourceHome;
import org.apache.ws.resource.faults.FaultException;
import org.apache.ws.resource.i18n.Keys;
import org.apache.ws.resource.i18n.MessagesImpl;
import org.apache.ws.util.i18n.Messages;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.xml.rpc.handler.MessageContext;
import javax.xml.rpc.handler.soap.SOAPMessageContext;
import javax.xml.soap.SOAPMessage;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * An abstract implementation of a resource context.
 */
public abstract class AbstractResourceContext
   implements ResourceContext
{
   private static final Log      LOG           = LogFactory.getLog( AbstractResourceContext.class.getName(  ) );
   private static final Messages MSG           = MessagesImpl.getInstance(  );
   private SOAPMessage           m_msg;
   private ResourceHome          m_home;
   private String                m_serviceName;
   private URL                   m_serviceURL;
   private Map                   m_propsMap    = Collections.synchronizedMap( new HashMap(  ) );
   private String                m_baseURL;
   private String                m_action;
   private SOAPMessageContext    m_msgContext;

   /**
    * Creates a new {@link AbstractResourceContext} object.
    *
    * @param soapMsgContext DOCUMENT_ME
    */
   public AbstractResourceContext( SOAPMessageContext soapMsgContext )
   throws Exception
   {
      if ( soapMsgContext == null )
      {
         throw new IllegalArgumentException( MSG.getMessage( Keys.NULL_SOAPMSGCTX ) );
      }

      m_msgContext    = soapMsgContext;
      m_msg           = soapMsgContext.getMessage(  );
      extractFields( soapMsgContext );
      try
      {
         m_home = (ResourceHome) new InitialContext(  ).lookup( getResourceHomeLocation(  ) );
      }
      catch ( NamingException ne )
      {
         ne.printStackTrace(  );
         throw new ResourceContextException( ne );
      }

      if ( m_home instanceof AbstractResourceHome )
      {
         AbstractResourceHome abstractResourceHome = ( (AbstractResourceHome) m_home );

         //check if home has been initialized
         synchronized ( abstractResourceHome )
         {
            if ( !abstractResourceHome.isInitialized(  ) )
            {
               abstractResourceHome.init(  );
            }
         }
      }
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String getBaseURL(  )
   {
      return m_baseURL;
   }

   /**
    * DOCUMENT_ME
    *
    * @param name  DOCUMENT_ME
    * @param value DOCUMENT_ME
    */
   public void setProperty( String name,
                            Object value )
   {
      m_propsMap.put( name, value );
   }

   /**
    * DOCUMENT_ME
    *
    * @param name DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public Object getProperty( String name )
   {
      return this.m_propsMap.get( name );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public Iterator getPropertyNames(  )
   {
      return this.m_propsMap.keySet(  ).iterator(  );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String getRequestAction(  )
   {
      return m_action;
   }

   /**
    * @see org.apache.ws.resource.ResourceContext#getResource()
    */
   public Resource getResource(  )
   throws ResourceException
   {
      ResourceHome home = getResourceHome(  );
      Object       id = home.extractResourceIdentifier( this );
      return home.find( id );
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public ResourceHome getResourceHome(  )
   {
      return m_home;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public synchronized String getResourceHomeLocation(  )
   {
      String homeLocation =
         JndiConstants.CONTEXT_NAME_SERVICES + "/" + getServiceName(  ) + "/" + JndiConstants.ATOMIC_NAME_HOME;
      LOG.debug( MSG.getMessage( Keys.JNDI_HOME_LOCATION, homeLocation ) );
      return homeLocation;
   }

   /**
    * @see ResourceContext#setResponseAction(java.net.URI)
    */
   public void setResponseAction( URI action )
   {
      if ( action != null )
      {
         AddressingHeaders responseAddrHeaders = AddressingUtils.getResponseHeaders( m_msgContext );
         try
         {
            responseAddrHeaders.setAction( action.toString(  ) );
         }
         catch ( Exception e )
         {
            throw new RuntimeException( "Failed to set wsa:Action for response.", e );
         }
      }
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public SOAPMessage getSOAPMessage(  )
   {
      return m_msg;
   }

   /**
    * Returns target m_serviceName associated with this context.
    *
    * @return the target m_serviceName that was set with {@link #setServiceName(String) setService()}. If set to
    *         <code>null</code> or not set at all, by default it returns the target m_serviceName associated with the
    *         underlying SOAP m_msg context.
    */
   public String getServiceName(  )
   {
      return m_serviceName;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public URL getServiceURL(  )
   {
      return m_serviceURL;
   }

   /**
    * DOCUMENT_ME
    *
    * @param name DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public boolean containsProperty( String name )
   {
      return this.m_propsMap.containsKey( name );
   }

   /**
    * DOCUMENT_ME
    *
    * @param name DOCUMENT_ME
    */
   public void removeProperty( String name )
   {
      this.m_propsMap.remove( name );
   }

   /**
    * Returns the WS-Addressing Action Header
    *
    * @param msgContext SAAJ message context; may not be null
    *
    * @return
    */
   protected String getAddressingAction( MessageContext msgContext )
   {
      AddressingHeaders addrHeaders =
         (AddressingHeaders) msgContext.getProperty( org.apache.axis.message.addressing.Constants.ENV_ADDRESSING_REQUEST_HEADERS );
      if ( addrHeaders == null )
      {
         LOG.debug( MSG.getMessage( Keys.NO_WSA_HEADERS ) );
         throw new FaultException( Soap1_1Constants.FAULT_CLIENT, "This endpoint requires WS-Addressing headers." );
      }

      Action wsaAction = addrHeaders.getAction(  );
      if ( wsaAction == null )
      {
         LOG.debug( MSG.getMessage( Keys.NO_WSA_ACTION ) );
         throw new FaultException( Soap1_1Constants.FAULT_CLIENT,
                                   "The WS-Addressing Action header is required by this endpoint." );
      }

      LOG.debug( MSG.getMessage( Keys.GET_ACTION_FROM_ADDR_HDR, wsaAction ) );
      return wsaAction.toString(  );
   }

   /**
    * Sets the target m_serviceName of this context.
    *
    * @param serviceName m_serviceName name.
    */
   protected void setServiceName( String serviceName )
   {
      m_serviceName = serviceName;
   }

   /**
    * DOCUMENT_ME
    *
    * @param msgContext DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   protected String getServiceName( MessageContext msgContext )
   {
      URL serviceURL = ( m_serviceURL != null ) ? m_serviceURL : getServiceURL(  );
      return serviceURL.getFile(  );
   }

   /**
    * DOCUMENT_ME
    *
    * @param msgContext SAAJ message context; may not be null
    *
    * @return DOCUMENT_ME
    */
   protected URL getServiceURL( MessageContext msgContext )
   {
      AddressingHeaders wsaHeaders =
         (AddressingHeaders) msgContext.getProperty( org.apache.axis.message.addressing.Constants.ENV_ADDRESSING_REQUEST_HEADERS );
      try
      {
         String serviceUrl = wsaHeaders.getTo(  ).toString(  );
         LOG.debug( MSG.getMessage( Keys.GET_SERVICE_URL_FROM_ADDR_HDR, serviceUrl ) );
         return new URL( serviceUrl );
      }
      catch ( MalformedURLException murle )
      {
         throw new RuntimeException( "Value of wsa:To SOAP header element is not a valid URL.", murle );
      }
   }

   private String getBaseURL( URL serviceURL )
   {
      String s = serviceURL.toString(  );
      s = s.substring( 0,
                       s.lastIndexOf( "/" ) );
      return s;
   }

   private void extractFields( MessageContext msgContext )
   {
      m_serviceURL     = getServiceURL( msgContext );
      m_serviceName    = getServiceName( msgContext );
      m_baseURL        = getBaseURL( m_serviceURL );
      m_action         = getAddressingAction( msgContext );
      extractProperties( msgContext );
   }

   private void extractProperties( MessageContext msgContext )
   {
      Iterator propertyNames = msgContext.getPropertyNames(  );
      while ( propertyNames.hasNext(  ) )
      {
         String keyName = (String) propertyNames.next(  );
         m_propsMap.put( keyName,
                         msgContext.getProperty( keyName ) );
      }
   }
}